#include <math.h>
#include <eyebot.h>

/**
 * Fuzzy controller for the velocity of the EyeBot
 */

#define ERROR_WIDTH 		50
#define ERRORDOT_WIDTH	 	30
#define MAX_CHANGE_SIZE		5

/**
 * {too_slow, speed_OK, too_fast} = v[3]
 * {getting_slower, no_accel, getting_faster} = a[3]
 * {go_slower, do_nothing, go_faster} = o[3]
 */

float v[3] = {0,0,0};
float a[3] = {0,0,0};
float o[3] = {0,0,0};

float centroid = 0;

void memberships(float, float*, int);

/*
 * Linguistic rules:
 * IF too_slow AND getting_slower THEN go_faster
 * IF speed_OK AND getting_slower THEN go_faster
 * IF too_fast AND getting_slower THEN go_slower
 *
 * IF too_slow AND no_accel THEN go_faster
 * IF speed_OK AND no_accel THEN do_nothing
 * IF too_fast AND no_accel THEN go_slower
 *
 * IF too_slow AND getting_faster THEN go_faster
 * IF speed_OK AND getting_faster THEN go_slower
 * IF too_fast AND getting_faster THEN go_slower
 */

float min(float a, float b) {
	if(a<b) return a;

 	return b;
}

int getFuzzyOutput(float error, float errorDot) {
	int i,j =0;
	int output = 0;

	float **x;
	//Obtain fuzzy set memberships from inputs
	memberships(error, (float *)v, ERROR_WIDTH);
	memberships(errorDot, (float *)a, ERRORDOT_WIDTH);

	a[0]=0;
	a[1]=1;
	a[2]=0;

	//Apply inferences, using ROOT-SUM-SQUARE output combination
	LCDPrintf("v: %1.1f %1.1f %1.1f\n",v[0],v[1],v[2]);
	LCDPrintf("a: %1.1f %1.1f %1.1f\n",a[0],a[1],a[2]);

	//go_faster
	o[2] = sqrt(min(v[0],a[0])*min(v[0],a[0]) +
			min(v[1],a[0])*min(v[1],a[0]) +
			min(v[0],a[1])*min(v[0],a[1]) +
			min(v[0],a[2])*min(v[0],a[2]));

	//go_slower
	o[0] = sqrt(min(v[2],a[0])*min(v[2],a[0]) +
			min(v[2],a[1])*min(v[2],a[1]) +
			min(v[1],a[2])*min(v[1],a[2]) +
			min(v[2],a[2])*min(v[2],a[2]));

	//do_nothing
	o[1] = min(v[1],a[1]);


	LCDPrintf("o: %1.1f %1.1f %1.1f\n",o[0],o[1],o[2]);


	//Calculate fuzzy centroid
	//= sum(center*strength)/sum(strength)
	//Here centre of o[0] = -100, centre of o[1] = 0, centre of o[2] = 100

	centroid = (-MAX_CHANGE_SIZE*o[0]+MAX_CHANGE_SIZE*o[2])/(o[0]+o[1]+o[2]);

	output = (int)centroid;
//	LCDPrintf("O: %d\n",output);

	//Ensure clipping to appropriate range
	output = (output < -MAX_CHANGE_SIZE)?(-MAX_CHANGE_SIZE):output;
	output = (output > MAX_CHANGE_SIZE)?(MAX_CHANGE_SIZE):output;

//	LCDPrintf("O: %d\n",output);

	return output;

}

//Takes error and returns partial memberships for linguistic variables
void memberships(float error, float *x, int error_width) {
	if (error < -error_width) {	
		x[0] = 1.0;
		x[1] = 0.0;
		x[2] = 0.0;
	}
	else if (error > -error_width && error < 0) {
		x[1] = ((float)(error + error_width))/error_width;
		x[0] = 1 - x[1];
		x[2] = 0.0;
	}
	else if (error == 0) {
		x[0] = 0.0;
		x[1] = 1.0;
		x[2] = 0.0;
	}
	else if (error > 0 && error < error_width) {
		x[0] = 0.0;
		x[2] = ((float)(error_width - error))/error_width;
		x[1] = 1 - x[2];
	}
	else if (error > error_width) {
		x[0] = 0.0;
		x[1] = 0.0;
		x[2] = 1.0;
	}
}
